//	msg.h
//	
//============================================

#ifndef MSG_H__
#define MSG_H__ 1


//#define STL_STYLE	1	/* uncomment this line for using STL containers */

#include <typeinfo.h>

#ifdef STL_STYLE
#include <map.h>
#include <bstring.h>
#else
#include <afx.h>
#include <afxtempl.h>
#endif


class Msg
{
public:
	Msg(){};
	virtual ~Msg() {};
};

class HandlerBase;

//	the base class for message target
//	contains virtual table for dispatchering
//============================================
class Target 
{

public:
	void Notify (const Msg& msg);

protected:
	Target() {};
	virtual ~Target() {};
	// Vtable for double dispatchered calls
	//============================================
#ifdef STL_STYLE
	typedef less<string> compare;
	class Container : protected map<string,HandlerBase*,compare>
	{
	public:
		Container() {};
		HandlerBase*	Find(const string& classname);
		void Insert (const string& classname, HandlerBase* handler);
		void Erase  (const string& classname);
	};
#else
	class Container : protected CTypedPtrMap<CMapStringToPtr, CString, HandlerBase*>
	{
	public:
		Container() {};
		HandlerBase*	Find(LPCTSTR classname);
		void Insert (LPCTSTR classname, HandlerBase* handler);
		void Erase  (LPCTSTR classname);
		~Container() { RemoveAll();};
	};
#endif

	Container Vtable;

};

// auxilary abstract class for double dispatchering
class HandlerBase
{
protected:
	friend class Target;
	virtual void IndirectCall (const Msg& msg)=0;

};


// template class for message handlers
//
// all message handlers have one copy of base 
// class Target and multiple copies of HandlerBase for each 
//==============================================================
template <class Message> class Handler :  
	virtual public Target, private HandlerBase
{
public:
	Handler();
	virtual ~Handler();

	virtual void OnNotify (const Message&)=0;

protected:
		// function for double dispatchering
	virtual void IndirectCall (const Msg& msg)
	{
		OnNotify ((const Message&)msg);
	};
private:
#ifdef STL_STYLE
	string	mMessageClass;
#else
	CString	mMessageClass;
#endif

};

template <class Message>
Handler<Message>::Handler()
{
		// here defines personal entry to Vtable accordingly to 
		// message class name and own copy of HandlerBase pointer
	Vtable.Insert (mMessageClass=typeid(Message).raw_name(),this);
};

template <class Message>
Handler<Message>::~Handler()
{		// remove Vtable entry
	Vtable.Erase(mMessageClass);
};


#endif
